# -*- coding: utf-8 -*-

import logging
import os
import sqlite3
import subprocess

from msgbus import TK_POS_GETPOSLIST, TK_SYS_NAK, FM_PARAM
from persistence import Driver

logger = logging.getLogger("Maintenance")


class CleanDatabasesProcessor(object):
    def __init__(self, mbcontext, comp_no, days_to_keep_orders):
        self.mbcontext = mbcontext
        self.comp_no = comp_no
        self.days_to_keep_orders = days_to_keep_orders

        self.maintenance_processor = None
        self.pos_list = None

    def set_server_parameters(self, maintenance_processor):
        self.maintenance_processor = maintenance_processor

    def clean_databases(self):
        try:
            # Server
            if self.comp_no == 0:
                self.pos_list = self._get_pos_list()

                self._clean_orders_database()
                self._clean_fiscal_database()
            # POS
            else:
                return
        except Exception as _:
            logger.exception("Error cleaning databases")

    def vacuum_databases(self):
        try:
            # Server
            if self.comp_no == 0:
                self.pos_list = self._get_pos_list()

                self._vacuum_production_database()
                self._vacuum_fiscal_database()
                self._vacuum_orders_database()
            # POS
            else:
                return
        except Exception as _:
            logger.exception("Error doing vacuum on databases")

    def _clean_orders_database(self):
        logger.info("Starting [_clean_orders_database]")
        try:
            orders_stored_all_pos = []
            conn = None

            for pos_id in sorted(self.pos_list, key=lambda _: int(_)):
                try:
                    conn = Driver().open(self.mbcontext, service_name='Persistence', dbname=str(pos_id))

                    logger.info('Starting backup process for POS: %s' % pos_id)
                    try:
                        orders_stored = self._backup_orders(conn)
                        if orders_stored:
                            orders_stored_all_pos.extend(orders_stored)
                        else:
                            logger.info("Have no orders to clean in order db: {}".format(pos_id))
                    except Exception as _:
                        logger.exception("Error doing backup - POS: %s" % pos_id)
                        break
                finally:
                    if conn:
                        conn.close()

            if orders_stored_all_pos:
                for pos_id in self.pos_list:
                    try:
                        conn = Driver().open(self.mbcontext, service_name='Persistence', dbname=str(pos_id))
                        orders_stored_list = [orders_stored_all_pos[x:x + 100] for x in xrange(0, len(orders_stored_all_pos), 100)]
                        logger.info("Starting [_delete_orders] por POS: {}".format(pos_id))
                        for orders_stored in orders_stored_list:
                            self._delete_orders(conn, orders_stored)
                        logger.info("Finished [_delete_orders] por POS: {}".format(pos_id))
                    finally:
                        if conn:
                            conn.close()

            logger.info('Finished [_clean_orders_database]')
        except Exception as _:
            logger.exception("Error cleaning order dbs")

    def _vacuum_orders_database(self):
        # For each POS - Backup Orders DB
        logger.info("Starting [_vacuum_orders_database]")
        try:
            order_db_files = []
            conn = None

            for pos_id in sorted(self.pos_list, key=lambda _: int(_)):
                try:
                    conn = Driver().open(self.mbcontext, service_name='Persistence', dbname=str(pos_id))

                    order_db_files.append(self.get_db_file(conn, "orderdb"))
                finally:
                    if conn:
                        conn.close()

            for order_db in order_db_files:
                conn = None
                old_file_size = os.path.getsize(order_db)/1024/1024
                logger.info('Doing VACUUM for order db: {}'.format(order_db.split('db')[1]))

                try:
                    conn = sqlite3.connect(order_db)
                    conn.execute("VACUUM;")  # This will decrease the size of the databases
                except Exception as ex:
                    logger.exception("Error doing vacuum - %s - Message %s" % (order_db, str(ex)))
                    if conn is not None:
                        conn.close()
                else:
                    conn.close()

                new_file_size = os.path.getsize(order_db)/1024/1024

                logger.info('Order db file sizes. OLD: {}mb; NEW: {}mb'.format(old_file_size, new_file_size))

            logger.info('Finished [_vacuum_orders_database]')
        except Exception as _:
            logger.exception("Error doing vacuum on order dbs")

    @staticmethod
    def _delete_orders(conn, orders_stored):
        purge_sql = '''
                         DELETE FROM orderdb.OrderCustomProperties WHERE OrderId IN ({0});
                         DELETE FROM orderdb.OrderItemCustomProperties WHERE OrderId IN ({0});
                         DELETE FROM orderdb.OrderStateHistory WHERE OrderId IN ({0});
                         DELETE FROM orderdb.OrderItem WHERE OrderId IN ({0});
                         DELETE FROM orderdb.OrderTax WHERE OrderId IN ({0});
                         DELETE FROM orderdb.OrderDiscount WHERE OrderId IN ({0});
                         DELETE FROM orderdb.OrderVoidHistory WHERE OrderId IN ({0});
                         DELETE FROM orderdb.OrderTender WHERE OrderId IN ({0});
                         DELETE FROM orderdb.OrderComment WHERE OrderId IN ({0});
                         DELETE FROM orderdb.Orders WHERE OrderId IN ({0});
                     '''.format(','.join(orders_stored))

        try:
            execute_db_transaction(conn, purge_sql, close_conn=False)
        except Exception as _:
            logger.exception("Error purging database - ROLLBACK")

    def _backup_orders(self, conn):
        logger.info("Starting [_backup_orders]")
        orders_stored = []
        cursor = conn.select('''SELECT OrderId FROM orderdb.Orders
                               WHERE BusinessPeriod < strftime('%Y%m%d', 'now', '-{0} days')
                               AND StateId IN (SELECT StateId FROM orderdb.OrderState WHERE StateDescr IN ('PAID', 'VOIDED'))
                               ORDER BY OrderId LIMIT 500'''.format(self.days_to_keep_orders))
        order_list = map(lambda x: x.get_entry('OrderId'), cursor)

        if len(order_list) < 1:
            return

        backup_sql = '''
                         INSERT OR REPLACE INTO backuporders.Orders SELECT * FROM orderdb.Orders WHERE OrderId IN ({0});
                         INSERT OR REPLACE INTO backuporders.OrderCustomProperties SELECT * FROM orderdb.OrderCustomProperties WHERE OrderId IN ({0});
                         INSERT OR REPLACE INTO backuporders.OrderItemCustomProperties SELECT * FROM orderdb.OrderItemCustomProperties WHERE OrderId IN ({0});
                         INSERT OR REPLACE INTO backuporders.OrderStateHistory SELECT * FROM orderdb.OrderStateHistory WHERE OrderId IN ({0});
                         INSERT OR REPLACE INTO backuporders.OrderItem SELECT * FROM orderdb.OrderItem WHERE OrderId IN ({0});
                         INSERT OR REPLACE INTO backuporders.OrderTax SELECT * FROM orderdb.OrderTax WHERE OrderId IN ({0});
                         INSERT OR REPLACE INTO backuporders.OrderDiscount SELECT * FROM orderdb.OrderDiscount WHERE OrderId IN ({0});
                         INSERT OR REPLACE INTO backuporders.OrderVoidHistory SELECT * FROM orderdb.OrderVoidHistory WHERE OrderId IN ({0});
                         INSERT OR REPLACE INTO backuporders.OrderTender SELECT * FROM orderdb.OrderTender WHERE OrderId IN ({0});
                         INSERT OR REPLACE INTO backuporders.OrderComment SELECT * FROM orderdb.OrderComment WHERE OrderId IN ({0});
                     '''.format(','.join(order_list))
        try:
            execute_db_transaction(conn, backup_sql, close_conn=False)
            orders_stored = order_list
            logger.info("Finished [_backup_orders]")
        except Exception as _:
            logger.exception("Error doing backup - ROLLBACK")
        return orders_stored

    def _clean_fiscal_database(self):
        try:
            logger.info("Starting [_clean_fiscal_database]")

            conn = None
            try:
                conn = Driver().open(self.mbcontext, service_name="FiscalPersistence")

                logger.info('Starting purge of orders on fiscal database')
                orders_backup = self._backup_fiscal(conn)
                if orders_backup:
                    self._delete_fiscal(conn, orders_backup)
                else:
                    logger.info("Have no order to clean in fiscal db")
            except Exception as ex:
                if "malformed" in ex.message:
                    logger.exception("Malformed database. Executing hard clean")
                    db_file = self.get_db_file(conn, "fiscal")
                    self._hard_clean_fiscal_database(db_file)
                logger.exception('Error purging orders on fiscal database')
                raise ex
            else:
                logger.info('Finalized purge of orders on fiscal database')
            finally:
                if conn:
                    conn.close()

            logger.info("Finished [_clean_fiscal_database]")
        except Exception as _:
            logger.exception("Error cleaning fiscal db")

    def _vacuum_fiscal_database(self):
        try:
            logger.info("Starting [_vacuum_fiscal_database]")

            conn = None
            try:
                conn = Driver().open(self.mbcontext, service_name="FiscalPersistence")
                db_file = self.get_db_file(conn, "fiscal")
            except Exception as ex:
                if "malformed" in ex.message:
                    logger.exception("Malformed database. Executing hard clean")
                    db_file = self.get_db_file(conn, "fiscal")
                    self._hard_clean_fiscal_database(db_file)
                logger.exception('Error doing vacuum on fiscal database')
                raise ex
            finally:
                if conn:
                    conn.close()

            old_file_size = os.path.getsize(db_file) / 1024 / 1024
            conn = None
            try:
                logger.info('Doing VACUUM for fiscal database')
                if db_file:
                    conn = sqlite3.connect(db_file)
                    conn.execute("VACUUM;")
            except Exception as _:
                logger.exception('Error doing vacuum on fiscal db')
            finally:
                if conn:
                    conn.close()

            new_file_size = os.path.getsize(db_file) / 1024 / 1024

            logger.info('Fiscal db file sizes. OLD: {}mb; NEW: {}mb'.format(old_file_size, new_file_size))

            logger.info("Finished [_vacuum_fiscal_database]")
        except Exception as _:
            logger.exception("Error doing vacuum on fiscal db")

    def _vacuum_production_database(self):
        try:
            logger.info("Starting [_vacuum_production_database]")

            conn = None
            try:
                conn = Driver().open(self.mbcontext)
                db_file = self.get_db_file(conn, "production")
            finally:
                if conn:
                    conn.close()

            old_file_size = os.path.getsize(db_file) / 1024 / 1024
            conn = None
            try:
                logger.info('Doing VACUUM on production database')
                if db_file:
                    conn = sqlite3.connect(db_file)
                    conn.execute("VACUUM;")
            except Exception as _:
                logger.exception('Error doing vacuum on production database')
            finally:
                if conn:
                    conn.close()

            new_file_size = os.path.getsize(db_file) / 1024 / 1024

            logger.info('Production database file sizes. OLD: {}mb; NEW: {}mb'.format(old_file_size, new_file_size))

            logger.info("Finished [_vacuum_production_database]")
        except Exception as _:
            logger.exception("Error doing vacuum on production database")

    @staticmethod
    def get_db_file(conn, db_name_to_find):
        db_file = None
        pragma_cursor = conn.select("PRAGMA database_list;")
        for pragma_row in pragma_cursor:
            db_name, db_file = map(pragma_row.get_entry, (1, 2,))
            if db_name == db_name_to_find:
                break
        return db_file

    @staticmethod
    def _delete_fiscal(conn, orders_backup):
        logger.info("Starting [delete_fiscal]")
        purge_sql = """DELETE FROM fiscal.FiscalData WHERE OrderId in ('{0}');
                       DELETE FROM fiscal.PaymentData WHERE OrderId in ('{0}');"""
        try:
            execute_db_transaction(conn, purge_sql.format("','".join(orders_backup)), close_conn=False)
            logger.info("Finished [delete_fiscal]")
        except Exception as ex:
            logger.exception("Error purging order {0} on fiscal database")
            if "malformed" in ex.message:
                raise ex

    def _backup_fiscal(self, conn):
        logger.info("Starting [_backup_fiscal]")
        cursor = conn.select('''SELECT OrderId
                                FROM fiscal.FiscalData
                                WHERE SentToBKC = 1 AND SentToBKOffice = 1 AND SentToNfce = 1 AND strftime('%s', 'now', '-{0} days') > DataNota
                                ORDER BY OrderId LIMIT 500'''.format(self.days_to_keep_orders))
        order_list = map(lambda x: x.get_entry('OrderId'), cursor)
        if len(order_list) < 1:
            return

        backup_sql = """INSERT OR REPLACE INTO backupfiscal.PaymentData SELECT * FROM fiscal.PaymentData WHERE OrderId in ('{0}');
                        INSERT OR REPLACE INTO backupfiscal.FiscalData SELECT * FROM fiscal.FiscalData WHERE OrderId in ('{0}');"""

        execute_db_transaction(conn, backup_sql.format("','".join(order_list)), close_conn=False)
        logger.info("Finished [_backup_fiscal]")
        return order_list

    def _get_pos_list(self):
        try:
            pos_list = []
            msg = self.mbcontext.MB_EasySendMessage('PosController', TK_POS_GETPOSLIST, FM_PARAM)
            if msg.token == TK_SYS_NAK or not msg.data:
                raise
            for pos_id in msg.data.split('\0'):
                if pos_id not in pos_list:
                    pos_list.append(pos_id)
            return sorted(pos_list)
        except Exception:
            logger.exception("Error getting the list of POS")

    @staticmethod
    def _hard_clean_fiscal_database(db_file):
        logger.info("Starting [_hard_clean_fiscal_database]")
        subprocess.call(["sqlite3",
                         db_file,
                         ".mode insert",
                         ".output dump_all.sql",
                         ".dump",
                         ".read dump_all.sql"], shell=True)
        logger.info("Finished [_hard_clean_fiscal_database]")


def execute_db_transaction(conn, query, close_conn=True):
    try:
        conn.transaction_start()
        conn.query('''BEGIN TRANSACTION;''')
        conn.query(query)
    except Exception as ex:
        conn.query('''ROLLBACK TRANSACTION;''')
        raise ex
    else:
        conn.query('''COMMIT TRANSACTION;''')
    finally:
        if close_conn:
            conn.close()



from logging import getLogger
from typing import Optional, Dict, Any, List

from nperipherals_core.model.custom_exceptions import DeviceNotConnected
from nperipherals_core.model.device import Device, WriterDevice

logger = getLogger(__name__)

try:
    import pywin32_bootstrap
    # noinspection PyUnresolvedReferences
    import win32gui
    # noinspection PyUnresolvedReferences
    import win32print
    # noinspection PyUnresolvedReferences
    import win32ui
    # noinspection PyUnresolvedReferences
    import win32con
except ImportError:
    logger.warning("Error importing python win32 libs", exc_info=True)


class WindowsApiPrinterDevice(WriterDevice, Device):
    FONTSIZE = 11
    HORIZONTAL_RESOLUTION = 8
    VERTICAL_RESOLUTION = 10
    LOG_PIXELS_X = 88
    LOG_PIXELS_Y = 90
    PHYSICAL_WIDTH = 110
    PHYSICAL_HEIGHT = 111
    scale_factor = 20
    printer_dict = None

    # noinspection SpellCheckingInspection
    paper_sizes = {
        "letter": 1,
        "lettersmall": 2,
        "tabloid": 3,
        "ledger": 4,
        "legal": 5,
        "statement": 6,
        "executive": 7,
        "a3": 8,
        "a4": 9,
        "envelope9": 19,
        "envelope10": 20,
        "envelope11": 21,
        "envelope12": 22,
        "envelope14": 23,
        "fanfold": 39,
    }

    orientations = {
        "portrait": 1,
        "landscape": 2,
    }

    duplexes = {
        "normal": 1,
        "none": 1,
        "long": 2,
        "short": 3,
    }

    def __init__(self, device_configurations):
        # type: (Dict[unicode, Any]) -> None
        self.get_available_devices = None  # type: Optional[List]
        self.name = device_configurations["Name"]  # type: unicode
        self.paper_size = device_configurations.get("PaperSize", 0)  # type:  Optional[unicode]
        self.duplex = device_configurations.get("Duplex", None)  # type: Optional[unicode]
        self.orientation = device_configurations.get("Orientation", None)  # type:  Optional[unicode]
        self.write_retries = int(device_configurations.get("Retries", 1))  # type: int
        self.write_retry_sleep_in_seconds = float(device_configurations.get("RetriesSleepSeconds", 1.0))  # type: float

        self.dc = None
        self.font = None
        self.printer = None
        self.hdc = None
        self.pen = None

        self.page = 0
        self.printer_dict = {}

    def initialize(self):
        return

    def connect(self):
        return

    def disconnect(self):
        return

    def write(self, data):
        # type: (unicode) -> None
        try:
            data = data.splitlines()
            self.start_document()
            self.set_font("Courier New", self.FONTSIZE)
            for i in range(len(data)):
                self.text((72, 72 + i * self.FONTSIZE + 16), data[i])

            self.end_document()

            logger.debug("Device [{}] sent data [{}]".format(self.name, data))

        except Exception as e:
            logger.exception("Error getting socket data with error [{}]".format(repr(e)))

        raise DeviceNotConnected()

    def get_write_retries(self):
        # type: () -> int

        return self.write_retries

    def get_write_retry_sleep_in_seconds(self):
        # type: () -> float

        return self.write_retry_sleep_in_seconds

    def start_document(self, desc="POS:WindowsApiManager"):
        self.name = win32print.GetDefaultPrinter()
        self.printer = win32print.OpenPrinter(self.name)

        dev_mode = win32print.GetPrinter(self.printer, 8)["pDevMode"]

        if self.paper_size is not None:
            if type(self.paper_size) is int:
                dev_mode.PaperSize = self.paper_size
            else:
                dev_mode.PaperSize = self.paper_sizes[self.paper_size]
        if self.orientation is not None:
            dev_mode.Orientation = self.orientations[self.orientation]
        if self.duplex is not None:
            dev_mode.Duplex = self.duplexes[self.duplex]

        self.hdc = win32gui.CreateDC("WINSPOOL", self.name, dev_mode)
        self.dc = win32ui.CreateDCFromHandle(self.hdc)

        self.dc.SetMapMode(win32con.MM_TWIPS)  # hundredths of inches

        try:
            self.dc.StartDoc(desc)
        except win32ui.error as ex:
            logger.debug(ex)

        self.dc.SetBkMode(win32con.TRANSPARENT)
        self.pen = win32ui.CreatePen(0, int(self.scale_factor), 0L)
        self.dc.SelectObject(self.pen)
        win32gui.SetBkMode(self.hdc, 1)  # transparent
        self.page = 1

    def end_document(self):
        if self.page == 0:
            return
        self.dc.EndDoc()
        del self.dc

    def getsize(self):
        if self.page == 0:
            self.start_document()
        horizontal = float(self.dc.GetDeviceCaps(self.HORIZONTAL_RESOLUTION))
        vertical = float(self.dc.GetDeviceCaps(self.VERTICAL_RESOLUTION))
        width = horizontal * (72.0 / self.dc.GetDeviceCaps(self.LOG_PIXELS_X))
        height = vertical * (72.0 / self.dc.GetDeviceCaps(self.LOG_PIXELS_Y))
        return width, height

    def text(self, position, text):
        if self.page == 0:
            self.start_document()
        self.dc.TextOut(int(self.scale_factor * position[0]),
                        int(-1 * self.scale_factor * position[1]), text)

    def set_font(self, name, size, bold=None, italic=None):
        if self.page == 0:
            self.start_document()
        wt = 400
        if bold:
            wt = 700
        if italic:
            italic = 1
        else:
            italic = 0
        self.font = self.get_font(name, size, wt, italic)
        self.dc.SelectObject(self.font)

    def get_font(self, name, size, weight=400, italic=0):
        if italic:
            return win32ui.CreateFont({
                "name": name,
                "height": self.scale_factor * size,
                "weight": weight,
                "italic": italic,
            })
        else:
            return win32ui.CreateFont({
                "name": name,
                "height": self.scale_factor * size,
                "weight": int(weight),
            })

    def list_printers(self):
        dft = win32print.GetDefaultPrinter()
        printers = {'keys': None}
        if printers['keys'] is None:
            self.build_dict()
        keys = printers.keys()
        keys.sort()
        rc = [dft]
        for k in keys:
            if k != dft:
                rc.append(k)
        return rc

    def is_device_available(self):
        return self.name in self.list_printers()

    def get_name(self):
        # type: () -> unicode
        return self.name

    def get_printer_jobs_count(self):
        printer = win32print.OpenPrinter(self.name)
        jobs = win32print.EnumJobs(printer, 0, -1, 1)
        win32print.ClosePrinter(printer)

        return len(jobs)

    def build_dict(self):
        lst = win32print.EnumPrinters(
            win32print.PRINTER_ENUM_CONNECTIONS
            + win32print.PRINTER_ENUM_LOCAL)
        self.printer_dict = {}
        for flags, description, name, comment in lst:
            self.printer_dict[name] = {}
            self.printer_dict[name]["flags"] = flags
            self.printer_dict[name]["description"] = description
            self.printer_dict[name]["comment"] = comment



from mw_helper import convert_to_dict, ensure_list
from production.box._OrderChangerProductionBox import OrderChangerProductionBox
from production.model import Item
from typing import List


class ComboFilterBox(OrderChangerProductionBox):
    def __init__(self, name, sons, items_to_keep, logger=None):
        super(ComboFilterBox, self).__init__(name, sons, logger)
        self.items_to_keep = convert_to_dict(ensure_list(items_to_keep))

    def change_order(self, order):
        filtered_items = []  # type: List[Item]
        for item in order.items:
            self.handle_item(item, filtered_items, 0)
        order.items = filtered_items

        return order

    def handle_item(self, item, filtered_items, levels_to_subtract, parent_qty=None):
        if parent_qty is not None:
            item.qty *= parent_qty

        if item.item_type in ("PRODUCT", "CANADD"):
            if item.original_level is None:
                item.original_level = item.level
            item.level = int(item.level) - levels_to_subtract
            filtered_items.append(item)
        elif item.item_type in ("COMBO", "OPTION"):
            if item.item_type == "COMBO" and item.part_code in self.items_to_keep:
                filtered_items.append(item)
            elif item.item_type == "OPTION":
                filtered_items.append(item)
            else:
                for son in item.items:
                    son.seat = item.seat
                    son.manual_course = item.manual_course
                    self.handle_item(son, filtered_items, levels_to_subtract + 1, item.qty)



from application.model.benefit import ActionResult
from posot import OrderTaker
from typing import Dict

from application.model.order import Order
from application.model.transactions import AddBenefitToApplyUnitOfWork
from application.model.benefit import ActionTypes, Action
from application.model.conditions import Restrictions

from ._DiscountService import DiscountService
from ._OrderService import OrderService


class ActionService(object):
    def __init__(self, order_service, product_groups):
        # type: (OrderService, Dict) -> None

        self.order_service = order_service
        self.product_groups = product_groups

    def apply_action(self, action, burnt_items, uow, pos_ot, restriction, on_total):
        # type: (Action, Dict, AddBenefitToApplyUnitOfWork, OrderTaker, Restrictions, bool) -> ActionResult

        order = self.order_service.get_order(uow.pos_id, order_picture=uow.order.order_picture)
        discount_ids = Order.get_available_discount_ids(pos_ot)
        service = DiscountService(pos_ot, self.product_groups)
        result = ActionResult([], {}, [], {}, False, False)

        if order.order_state == "TOTALED":
            if action.name == ActionTypes.discountAmountOnTotal.value:
                result = service.apply_discount_amount_on_total(action, order, restriction, discount_ids)

            if action.name == ActionTypes.discountPercentageOnTotal.value:
                result = service.apply_discount_percentage_on_total(action, order, restriction, discount_ids)

            if action.name == ActionTypes.discountAmountByItems.value:
                result = service.apply_discount_amount_by_items(action, order, restriction, discount_ids, False)

            if action.name == ActionTypes.discountAmountByItemsOnTotal.value:
                result = service.apply_discount_amount_by_items(action, order, restriction, discount_ids, True)

            if action.name == ActionTypes.discountPercentageByItems.value:
                result = service.apply_discount_percentage_by_items(action, order, restriction, discount_ids, False)

            if action.name == ActionTypes.discountPercentageByGroupsOnTotal.value:
                result = service.apply_discount_percentage_by_groups(action, order, restriction, discount_ids, True)

            if action.name == ActionTypes.discountAmountByGroupsOnTotal.value:
                result = service.apply_discount_amount_by_groups(action, order, restriction, discount_ids, True)

            if action.name == ActionTypes.discountPercentageByItemsOnTotal.value:
                result = service.apply_discount_percentage_by_items(action, order, restriction, discount_ids, True)

            if action.name == ActionTypes.paymentAmount.value:
                result = service.apply_discount_amount_on_total_by_tender(action, order, restriction)

            if action.name == ActionTypes.paymentPercentage.value:
                result = service.apply_discount_percentage_on_total_by_tender(action, order, restriction)

            if action.name == ActionTypes.paymentAmountOnItem.value:
                result = service.apply_discount_amount_on_total_by_tender(action, order, restriction, True)

            if action.name == ActionTypes.itemsToAdd.value:
                result = service.apply_tender_on_added_items(action, order, uow.benefit_order_info)

            if result.benefit_applied:
                result.items_to_burn = order.get_items_to_burn(order, uow.benefit, burnt_items, self.product_groups)

            return result

        elif order.order_state == "IN_PROGRESS":
            if action.name == ActionTypes.itemsToAdd.value:
                return service.apply_items_to_add(action, burnt_items, uow, restriction)

            if on_total is False:
                result.items_to_burn = order.get_items_to_burn(order, uow.benefit, burnt_items, self.product_groups)
                result.pending_to_apply = True
                return result

        return ActionResult([], {}, [], {}, False, False)



from pos_model import SaleItem, TaxItem
from typing import List

from _TaxCalculator import TaxCalculator


class GeneralTaxCalculator(object):
    def __init__(self, tax_calculators):
        # type: (List[TaxCalculator]) -> None
        self.tax_calculators = tax_calculators

    def calculate_all_taxes(self, sale_item):
        # type: (SaleItem) -> List[TaxItem]
        previous_taxes = []
        for tax_calculator in self.tax_calculators:
            tax_item = tax_calculator.calculate_tax(sale_item, previous_taxes)
            if tax_item is not None:
                previous_taxes.append((sale_item, tax_item))

        calculates_taxes = []
        for tax_tuple in previous_taxes:
            calculates_taxes.append(tax_tuple[1])

        return calculates_taxes



import logging

from actions.models import PaymentType
from actions.coupon.apply_coupon import apply_coupon
from actions.discount.clean_discounts import clean_discounts_and_payments
from actions.discount.get_discount import get_visible_discounts
from application.interactor import NDiscountInteractor
from model.customexception import ErrorAddingBenefits
from mw_helper import show_filtered_list_box_dialog, show_numpad_dialog, show_message_dialog
from sysactions import show_messagebox, get_current_order, StopAction, get_model, get_posot, show_keyboard

logger = logging.getLogger("PosActions")


def apply_discount(pos_id):
    try:
        coupons, disconnected_discount_component = get_visible_discounts(pos_id)
        if len(coupons) == 0:
            show_messagebox(pos_id, "$DISCOUNT_NOT_FOUND", icon="warning")
            return

        selected_index = _select_discount(pos_id, coupons)
        if selected_index is None:
            return

        selected_coupon = filter(lambda c: c["index"] == selected_index + 1, coupons)
        if not selected_coupon:
            return

        selected_coupon = selected_coupon[0]
        benefit_id = selected_coupon.get("benefit_id", "")
        coupon = selected_coupon.get("coupon", {})

        if disconnected_discount_component:
            clean_discounts_and_payments(pos_id, disconnected_discount_component)
            model = get_model(pos_id)
            pos_ot = get_posot(model)
            order = get_current_order(model)
            apply_amount_discount(pos_id, pos_ot, order, selected_coupon)
        elif coupon.get("askValue", False):
            coupon["index"] = _get_discount_type(coupon)
            amount = _get_discount_amount(coupon, pos_id)
            if coupon.get("requestJustification", False):
                coupon["justification"] = _get_justification_discount(pos_id)

            return _set_amount_discount(pos_id, coupon, amount)
        else:
            return apply_coupon(pos_id, table_id=None, coupon_code=benefit_id)
    except ErrorAddingBenefits as ex:
        logger.warn("Error applying discount ${}".format(ex.message))
        show_message_dialog(pos_id, "$ERROR", "${}".format(ex.message))
    except Exception as ex:
        logger.exception("Error applying discount")
        show_message_dialog(pos_id, "$ERROR", "${}".format(ex.message))

    return False


def _get_discount_amount(coupon, pos_id):
    model = get_model(pos_id)
    order = get_current_order(model)
    amount, percentage = _ask_discount_amount(pos_id, order, coupon=coupon)
    if percentage is not None:
        amount = percentage

    return amount


def apply_amount_discount(pos_id, pos_ot, order, coupon=None, discount_type=None, discount_value=None):
    amount, _ = _ask_discount_amount(pos_id, order, coupon, discount_type, discount_value)
    pos_ot.applyDiscount(1, amount, forcebyitem=1)


def _ask_discount_amount(pos_id, order, coupon=None, discount_type=None, discount_value=None):
    if not discount_type:
        discount_type = coupon['index']
        title = "$DISCOUNT_VALUE_TITLE" if discount_type == PaymentType.AMOUNT.value else "$DISCOUNT_PERCENTAGE_TITLE"
        mask = "CURRENCY" if discount_type == PaymentType.AMOUNT.value else "NUMBER"
        if not discount_value:
            discount_value = show_numpad_dialog(pos_id, title, "", mask)
            if discount_value is None:
                raise StopAction()

    if discount_type == PaymentType.AMOUNT.value:
        return discount_value, None

    amount = float(order.get("totalGross"))

    return (amount * discount_value) / 100, discount_value


def _select_discount(pos_id, coupons):
    options = ["{0} - {1}".format(coupon["index"], coupon["name"].encode("UTF-8")) for coupon in coupons]
    return show_filtered_list_box_dialog(pos_id, options, "$SELECT_ONE_COUPON", mask="NOFILTER")


def _set_amount_discount(pos_id, coupon, amount):
    coupon = set_coupon_amount(coupon, amount)
    return NDiscountInteractor.add_benefit(pos_id, "", coupon)


def _get_object_json(json_object, key, value):
    return [obj for obj in json_object if obj[key] == value][0]


def set_coupon_amount(coupon, amount):
    if coupon.get("actions", False) and coupon.get("actions").get("0", False):
        action = coupon.get("actions").get("0")
        value_actions = ["discountAmountOnTotal", "discountPercentageOnTotal"]
        group_actions = ["discountPercentageByGroupsOnTotal", "discountAmountByGroupsOnTotal"]
        if any(x for x in value_actions if x in action.keys()):
            action_name = action.keys()[0]
            action[action_name] = amount
            return coupon
        elif any(x for x in group_actions if x in action.keys()):
            action.values()[0][0]["amount"] = amount
            return coupon

    raise StopAction()


def _get_discount_type(coupon):
    if coupon.get("actions", False) and coupon.get("actions").get("0", False):
        action = coupon.get("actions").get("0")
        amount_actions = ["discountAmountOnTotal", "discountAmountByGroupsOnTotal"]
        percentage_actions = ["discountPercentageOnTotal", "discountPercentageByGroupsOnTotal"]
        if any(x for x in amount_actions if x in action.keys()):
            return PaymentType.AMOUNT.value
        elif any(x for x in percentage_actions if x in action.keys()):
            return PaymentType.PERCENTAGE.value

    raise StopAction()


def _get_justification_discount(pos_id):
    # type: (str) -> str

    justification = None
    while justification in (None, "", ".") or len(justification) < 5:
        justification = show_keyboard(pos_id, "$JUSTIFICATION_DISCOUNT", '', timeout=1800000, buttons="$OK")
        if justification:
            return justification.strip()



from ._NetworkDevice import NetworkDevice
from ._NetworkScannerDevice import NetworkScannerDevice
from ._NetworkEscPosPrinterDevice import NetworkEscPosPrinterDevice

try:
    from ._WindowsApiPrinterDevice import WindowsApiPrinterDevice
except ImportError:
    class WindowsApiPrinterDevice(object):
        pass



import logging
import time
from typing import Optional

from msgbus import MBEasyContext, TK_PERSIST_SEQUENCER_INCREMENT, FM_PARAM, TK_SYS_NAK

from comp_exceptions import FiscalIncrementException
from persistence import Connection
from old_helper import BaseRepository
from sysactions import send_message

logger = logging.getLogger("FiscalWrapper")


class SequencerRepository(BaseRepository):
    def __init__(self, mb_context):
        # type: (MBEasyContext) -> None

        self.service_name = "FiscalPersistence"
        self.fiscal_id_name = "FiscalId"

        super(SequencerRepository, self).__init__(mb_context)

    def get_fiscal_id(self):
        # type: () -> Optional[int]

        def inner_func(connection):
            # type: (Connection) -> Optional[int]

            query = """ 
                SELECT SeqNo
                FROM Sequencer
                WHERE SeqNm = '{}'
            """.format(self.fiscal_id_name)

            cursor = connection.select(query)

            if cursor.rows() == 0:
                return None
            else:
                return int(cursor.get_row(0).get_entry(0))

        fiscal_id = self.execute_with_connection(inner_func, service_name=self.service_name)
        return fiscal_id

    def set_fiscal_id(self, fiscal_id):
        # type: (int) -> None

        def inner_func(connection):
            # type: (Connection) -> None

            query = """
                UPDATE Sequencer 
                SET SeqNo={} 
                WHERE SeqNm = '{}'
            """.format(fiscal_id, self.fiscal_id_name)
            connection.query(query)

        self.execute_with_connection(inner_func, service_name=self.service_name)

    def increment_fiscal_id(self):
        # type: () -> int

        error_msg = "Error incrementing mandatory sequence [{}] on [{}]".format(self.fiscal_id_name, self.service_name)

        for _ in range(0, 3):
            try:
                msg = send_message(self.service_name, TK_PERSIST_SEQUENCER_INCREMENT, FM_PARAM, self.fiscal_id_name)
                if msg.token == TK_SYS_NAK:
                    logger.warning(error_msg)
                    time.sleep(0.1)
                    continue

                next_fiscal_id = int(msg.data.split('\0')[0])
                return next_fiscal_id

            except Exception as ex:
                logger.warning(error_msg, ex)
                continue

        raise FiscalIncrementException(error_msg)



import logging
import threading
from logging import Logger
from typing import Optional, List, Type

from messageprocessor import MessageProcessorCallback

default_logger = logging.getLogger("MessageProcessor.LoggingProcessorCallback")


class LoggingProcessorCallback(MessageProcessorCallback):
    def __init__(self, logger, ignored_exceptions=None):
        # type: (Logger, Optional[List[Type[Exception]]]) -> None
        self.logger = logger if logger is not None else default_logger
        self.ignored_exceptions = {}
        if ignored_exceptions is not None:
            for ex in ignored_exceptions:
                self.ignored_exceptions[ex] = ex
        self.local_storage = threading.local()

    def process_begun(self, event_name, request_id, data):
        # type: (str, str, bytes) -> None
        self.local_storage.original_thread_name = threading.current_thread().name
        self._set_thread_name(event_name, request_id)
        try:
            converted_data = data.decode(u"utf-8")
        except:
            try:
                converted_data = str(data)
            except:
                converted_data = u"could not convert data: {}. Implementing __str__ might help".format(type(data))

        self.logger.debug(u"Start handling. Data: {}".format(converted_data))

    def input_parsed(self, event_name, request_id, input_obj):
        self._set_thread_name(event_name, request_id)
        self.logger.debug(u"Input correctly parsed")

    def business_called(self, event_name, request_id, input_obj, output_obj):
        pass

    def process_finished(self, event_name, request_id, input_obj, output_obj):
        self.logger.info(u"Successfully processed")
        self._restore_thread_name()

    def parse_exception(self, event_name, request_id, data, exception):
        if type(exception) not in self.ignored_exceptions:
            self.logger.exception(u"Parse exception handling")
        self._restore_thread_name()

    def unhandled_exception(self, event_name, request_id, input_obj, exception):
        if type(exception) not in self.ignored_exceptions:
            self.logger.exception(u"Unhandled exception handling")
        self._restore_thread_name()

    @staticmethod
    def _set_thread_name(event_name, request_id):
        threading.current_thread().name = "MessageProcessorThread${}${}".format(event_name, request_id)

    def _restore_thread_name(self):
        threading.current_thread().name = self.local_storage.original_thread_name



from production.box import CFHHiderBox
from production.treebuilder import BoxConfiguration

from ._BoxBuilder import BoxBuilder


class CFHHiderBoxBuilder(BoxBuilder):
    def __init__(self, product_repository, loggers):
        super(CFHHiderBoxBuilder, self).__init__(loggers)
        self.product_repository = product_repository

    def build(self, config):
        # type: (BoxConfiguration) -> CFHHiderBox
        return CFHHiderBox(config.name,
                           config.sons,
                           self.product_repository.get_cfh_items(),
                           self.get_logger(config))



#
# The Python Imaging Library.
# $Id$
#
# standard channel operations
#
# History:
# 1996-03-24 fl   Created
# 1996-08-13 fl   Added logical operations (for "1" images)
# 2000-10-12 fl   Added offset method (from Image.py)
#
# Copyright (c) 1997-2000 by Secret Labs AB
# Copyright (c) 1996-2000 by Fredrik Lundh
#
# See the README file for information on usage and redistribution.
#

from . import Image


def constant(image, value):
    """Fill a channel with a given grey level.

    :rtype: :py:class:`~PIL.Image.Image`
    """

    return Image.new("L", image.size, value)


def duplicate(image):
    """Copy a channel. Alias for :py:meth:`PIL.Image.Image.copy`.

    :rtype: :py:class:`~PIL.Image.Image`
    """

    return image.copy()


def invert(image):
    """
    Invert an image (channel).

    .. code-block:: python

        out = MAX - image

    :rtype: :py:class:`~PIL.Image.Image`
    """

    image.load()
    return image._new(image.im.chop_invert())


def lighter(image1, image2):
    """
    Compares the two images, pixel by pixel, and returns a new image containing
    the lighter values. At least one of the images must have mode "1".

    .. code-block:: python

        out = max(image1, image2)

    :rtype: :py:class:`~PIL.Image.Image`
    """

    image1.load()
    image2.load()
    return image1._new(image1.im.chop_lighter(image2.im))


def darker(image1, image2):
    """
    Compares the two images, pixel by pixel, and returns a new image containing
    the darker values. At least one of the images must have mode "1".

    .. code-block:: python

        out = min(image1, image2)

    :rtype: :py:class:`~PIL.Image.Image`
    """

    image1.load()
    image2.load()
    return image1._new(image1.im.chop_darker(image2.im))


def difference(image1, image2):
    """
    Returns the absolute value of the pixel-by-pixel difference between the two
    images. At least one of the images must have mode "1".

    .. code-block:: python

        out = abs(image1 - image2)

    :rtype: :py:class:`~PIL.Image.Image`
    """

    image1.load()
    image2.load()
    return image1._new(image1.im.chop_difference(image2.im))


def multiply(image1, image2):
    """
    Superimposes two images on top of each other.

    If you multiply an image with a solid black image, the result is black. If
    you multiply with a solid white image, the image is unaffected. At least
    one of the images must have mode "1".

    .. code-block:: python

        out = image1 * image2 / MAX

    :rtype: :py:class:`~PIL.Image.Image`
    """

    image1.load()
    image2.load()
    return image1._new(image1.im.chop_multiply(image2.im))


def screen(image1, image2):
    """
    Superimposes two inverted images on top of each other. At least one of the
    images must have mode "1".

    .. code-block:: python

        out = MAX - ((MAX - image1) * (MAX - image2) / MAX)

    :rtype: :py:class:`~PIL.Image.Image`
    """

    image1.load()
    image2.load()
    return image1._new(image1.im.chop_screen(image2.im))


def add(image1, image2, scale=1.0, offset=0):
    """
    Adds two images, dividing the result by scale and adding the
    offset. If omitted, scale defaults to 1.0, and offset to 0.0.
    At least one of the images must have mode "1".

    .. code-block:: python

        out = ((image1 + image2) / scale + offset)

    :rtype: :py:class:`~PIL.Image.Image`
    """

    image1.load()
    image2.load()
    return image1._new(image1.im.chop_add(image2.im, scale, offset))


def subtract(image1, image2, scale=1.0, offset=0):
    """
    Subtracts two images, dividing the result by scale and adding the offset.
    If omitted, scale defaults to 1.0, and offset to 0.0. At least one of the
    images must have mode "1".

    .. code-block:: python

        out = ((image1 - image2) / scale + offset)

    :rtype: :py:class:`~PIL.Image.Image`
    """

    image1.load()
    image2.load()
    return image1._new(image1.im.chop_subtract(image2.im, scale, offset))


def add_modulo(image1, image2):
    """Add two images, without clipping the result. At least one of the images
    must have mode "1".

    .. code-block:: python

        out = ((image1 + image2) % MAX)

    :rtype: :py:class:`~PIL.Image.Image`
    """

    image1.load()
    image2.load()
    return image1._new(image1.im.chop_add_modulo(image2.im))


def subtract_modulo(image1, image2):
    """Subtract two images, without clipping the result. At least one of the
    images must have mode "1".

    .. code-block:: python

        out = ((image1 - image2) % MAX)

    :rtype: :py:class:`~PIL.Image.Image`
    """

    image1.load()
    image2.load()
    return image1._new(image1.im.chop_subtract_modulo(image2.im))


def logical_and(image1, image2):
    """Logical AND between two images. At least one of the images must have
    mode "1".

    .. code-block:: python

        out = ((image1 and image2) % MAX)

    :rtype: :py:class:`~PIL.Image.Image`
    """

    image1.load()
    image2.load()
    return image1._new(image1.im.chop_and(image2.im))


def logical_or(image1, image2):
    """Logical OR between two images. At least one of the images must have
    mode "1".

    .. code-block:: python

        out = ((image1 or image2) % MAX)

    :rtype: :py:class:`~PIL.Image.Image`
    """

    image1.load()
    image2.load()
    return image1._new(image1.im.chop_or(image2.im))


def logical_xor(image1, image2):
    """Logical XOR between two images. At least one of the images must have
    mode "1".

    .. code-block:: python

        out = ((bool(image1) != bool(image2)) % MAX)

    :rtype: :py:class:`~PIL.Image.Image`
    """

    image1.load()
    image2.load()
    return image1._new(image1.im.chop_xor(image2.im))


def blend(image1, image2, alpha):
    """Blend images using constant transparency weight. Alias for
    :py:meth:`PIL.Image.Image.blend`.

    :rtype: :py:class:`~PIL.Image.Image`
    """

    return Image.blend(image1, image2, alpha)


def composite(image1, image2, mask):
    """Create composite using transparency mask. Alias for
    :py:meth:`PIL.Image.Image.composite`.

    :rtype: :py:class:`~PIL.Image.Image`
    """

    return Image.composite(image1, image2, mask)


def offset(image, xoffset, yoffset=None):
    """Returns a copy of the image where data has been offset by the given
    distances. Data wraps around the edges. If **yoffset** is omitted, it
    is assumed to be equal to **xoffset**.

    :param xoffset: The horizontal distance.
    :param yoffset: The vertical distance.  If omitted, both
        distances are set to the same value.
    :rtype: :py:class:`~PIL.Image.Image`
    """

    if yoffset is None:
        yoffset = xoffset
    image.load()
    return image._new(image.im.offset(xoffset, yoffset))



# -*- coding: utf-8 -*-

# noinspection SpellCheckingInspection
class NfeCardBrand(object):
    # Tabela de bandeiras de Cartões de Crédito/Débito - Publicada em 16/03/2021
    # http://www.nfe.fazenda.gov.br/portal/listaConteudo.aspx?tipoConteudo=Iy/5Qol1YbE=
    # http://www.nfe.fazenda.gov.br/portal/exibirArquivo.aspx?conteudo=txtcX/WzVFY=

    visa = 1
    mastercard = 2
    american_express = 3
    sorocred = 4
    diners_club = 5
    elo = 6
    hipercard = 7
    aura = 8
    cabal = 9
    alelo = 10
    banes_card = 11
    cal_card = 12
    credz = 13
    discover = 14
    good_card = 15
    green_card = 16
    hiper = 17
    jc_b = 18
    mais = 19
    max_van = 20
    policard = 21
    rede_compras = 22
    sodexo = 23
    vale_card = 24
    verocheque = 25
    vr = 26
    ticket = 27
    outros = 99





class CashTransferSummaryInfo(object):
    def __init__(self, cash_out, pos_id, store_id, operator_id, business_date, value, envelope_number, authorizer):
        # type: (bool, int, int, int, unicode, float, int, unicode) -> ()
        self.cash_out = cash_out
        self.pos_id = pos_id
        self.store_id = store_id
        self.operator_id = operator_id
        self.business_date = business_date
        self.value = value
        self.envelope_number = envelope_number
        self.authorizer = authorizer



from helper import ExtendedEnum
from messagebus import TokenPriority, TokenCreator

MESSAGE_GROUP = "96"


class ListenedMessages(ExtendedEnum):
    TK_CUSTOMER_ADD_CUSTOMER = TokenCreator.create_token(TokenPriority.low, MESSAGE_GROUP, "1")
    TK_CUSTOMER_GET_CUSTOMER = TokenCreator.create_token(TokenPriority.low, MESSAGE_GROUP, "2")
    TK_CUSTOMER_UPDATE_CUSTOMER = TokenCreator.create_token(TokenPriority.low, MESSAGE_GROUP, "3")
    TK_CUSTOMER_DELETE_CUSTOMER = TokenCreator.create_token(TokenPriority.low, MESSAGE_GROUP, "4")



# -*- coding: utf-8 -*-

from manager.reports import opened_tables_report
from sysactions import show_messagebox
from pos_model import TableStatus
from xml.etree import cElementTree as eTree


def show_open_table_dialog(pos_id):
    ret = show_messagebox(pos_id, "$NEED_TO_CLOSE_ALL_TABLES_AND_TABS", buttons="$YES|$NO")
    if ret == 0:
        opened_tables_report(pos_id)


def has_tab_or_tables_opened(period, pos_id, operator_id=None):
    from msgbusboundary import MsgBusTableService as tableService
    from table_actions import has_table_service

    if not has_table_service():
        return False

    all_tables = tableService().list_tables(pos_id)

    for table in all_tables:
        if table.business_period != int(period):
            continue

        if not _is_open(table):
            continue

        if operator_id:
            table_picture = eTree.XML(tableService().get_table_picture(pos_id, table.id))
            table_user_id = table_picture.get("userId")
            if table_user_id != operator_id:
                continue
            return True

        return True

    return False


def _is_open(table):
    if "TAB" in table.id:
        is_open = int(table.status) not in [TableStatus.AVAILABLE, TableStatus.CLOSED]
    else:
        is_open = int(table.status) != TableStatus.AVAILABLE

    return is_open


def table_is_totaled(table):
    return table.status == TableStatus.TOTALIZED




from mw_helper import ensure_list
from production.box import ItemSorterBox
from production.treebuilder import BoxConfiguration

from ._BoxBuilder import BoxBuilder


class ItemSorterBoxBuilder(BoxBuilder):
    def __init__(self, product_repository, loggers):
        super(ItemSorterBoxBuilder, self).__init__(loggers)
        self.product_repository = product_repository
        self.loggers = loggers

    def build(self, config):
        # type: (BoxConfiguration) -> ItemSorterBox

        sort_order = ensure_list(self.get_extra(config, "SortOrder", []))
        item_property_sorters = ensure_list(self.get_extra(config, "ItemPropertySorters", []))

        return ItemSorterBox(
            config.name,
            config.sons,
            self.product_repository,
            sort_order,
            item_property_sorters,
            self.get_logger(config)
        )



import sys
import os
import cfgtools

from cfgtools import Group, Key
from msgbus import MBEasyContext
from messagehandler import MessageHandler
from eventhandler import KdsMonitorEventHandler
from kdsmonitor import KdsMonitor, Kds
from old_helper import config_logger

# COMMENT HERE
debugPath = '../python/pycharm-debug.egg'
if os.path.exists(debugPath):
    try:
        sys.path.index(debugPath)
    except:
        sys.path.append(debugPath)
    import pydevd

# Use the line below in the function you want to debug
# pydevd.settrace('localhost', port=9123, stdoutToServer=True, stderrToServer=True)
# UNTIL HERE

REQUIRED_SERVICES = "ProductionSystem|Persistence"


def main():
    config_logger(os.environ["LOADERCFG"], 'KdsMonitor')

    config = cfgtools.read(os.environ["LOADERCFG"])
    mbcontext = MBEasyContext("KdsMonitor")

    kds_list = []
    kds_group = config.find_group("KdsMonitor.KdsDict")  # type: Group
    for kds_key in kds_group.keys:  # type: Key
        name = kds_key.name
        ip = kds_key.descr
        view_name = str(kds_key.values[0])
        kds = Kds(name, ip, view_name)
        kds_list.append(kds)

    monitor_interval = int(config.find_value("KdsMonitor.MonitorInterval") or "15")
    max_error_count = int(config.find_value("KdsMonitor.MaxErrorCount") or "3")
    start_wait_time = int(config.find_value("KdsMonitor.StartWaitTime") or "120")

    kds_monitor_thread = KdsMonitor(mbcontext, kds_list, monitor_interval, max_error_count, start_wait_time)
    kds_monitor_thread.daemon = True

    kds_monitor_event_handler = KdsMonitorEventHandler(mbcontext, kds_monitor_thread)
    kds_message_handler = MessageHandler(mbcontext, "KdsMonitor", "KdsMonitor", REQUIRED_SERVICES, kds_monitor_event_handler)

    kds_monitor_thread.start()
    kds_message_handler.handle_events()



from escposformatter import CommandProcessor
from report.command import RepeatCommand


class RepeatCommandProcessor(CommandProcessor):
    def get_bytes(self, command):
        if not isinstance(command, RepeatCommand):
            return ""

        return command.text*command.count



from commons.dto import DefaultHeaderDto  # noqa
from _CashReportBodyDto import CashReportBodyDto  # noqa


class CashReportDto(object):
    def __init__(self, report_header_dto, report_body_dto, pos_id):
        # type: (DefaultHeaderDto, CashReportBodyDto, int) -> None
        self.report_header_dto = report_header_dto
        self.report_body_dto = report_body_dto
        self.pos_id = pos_id



import json
from abc import abstractmethod
from xml.etree import cElementTree as eTree

from application.model import NDiscountTokens
from application.model.configuration import Configurations
from application.model.customexception import ErrorApplyingBenefit
from mbcontextmessagehandler import MbContextMessageBus
from messagebus import Message, DataType
from msgbus import TK_SYS_NAK
from typing import Optional


class DefaultBenefitApplierRepository(object):

    def __init__(self, configs, message_bus):
        # type: (Configurations, MbContextMessageBus) -> None

        self.logger = configs.logger
        self.configs = configs
        self.message_bus = message_bus

    def add_benefit(self, current_pos_id, benefit_id, benefit, benefit_applier, tender_id):
        # type: (int, str, str, str, str) -> None

        data = json.dumps(dict(posId=current_pos_id, benefitId=benefit_id, benefit=benefit,
                               benefit_applier=benefit_applier, tender_id=tender_id))

        message = Message(token=NDiscountTokens.TK_NDISCOUNT_ADD_BENEFIT,
                          data=data,
                          data_type=DataType.string)

        response = self.message_bus.send_message("nDiscount", message)
        if response.token == TK_SYS_NAK:
            raise ErrorApplyingBenefit(response.data)

    def remove_benefit(self, current_pos_id, benefit_ids=None):
        # type: (int, Optional[str, list, None]) -> Message

        data = json.dumps(dict(pos_id=current_pos_id, benefit=benefit_ids))
        message = Message(token=NDiscountTokens.TK_NDISCOUNT_REMOVE,
                          data=data,
                          data_type=DataType.string)

        return self.message_bus.send_message("nDiscount", message)

    def apply_automatics_benefits(self, current_pos_id):
        # type: (str) -> Message

        message = Message(token=NDiscountTokens.TK_NDISCOUNT_APPLY_AUTOMATIC_BENEFITS,
                          data=json.dumps({"pos_id": current_pos_id}),
                          data_type=DataType.string)

        return self.message_bus.send_message("nDiscount", message)

    @abstractmethod
    def check_voucher(self, pos_id, benefit_id):
        # type: (int, str) -> None

        raise NotImplementedError()

    @abstractmethod
    def get_benefit_and_lock(self, benefit_id):
        # type: (str) -> str

        raise NotImplementedError()

    @abstractmethod
    def unlock_benefit(self, benefit_id):
        # type: (str) -> None

        raise NotImplementedError()

    @abstractmethod
    def burn_benefit(self, benefit_id):
        # type: (str) -> None

        raise NotImplementedError()

    @abstractmethod
    def get_benefit_info(self, data):
        # type: (str) -> None

        raise NotImplementedError()

    @abstractmethod
    def get_benefit_id_by_custom_property(self, order_picture):
        # type: (eTree) -> str

        raise NotImplementedError()



from abc import ABCMeta
from xml.etree import cElementTree as eTree

from dateutil import tz
from msgbus import TK_SYS_ACK, FM_PARAM, TK_HV_COMPLIST, MBEasyContext
from persistence import Connection, Driver
from sysactions import show_any_dialog, show_keyboard
from systools import sys_log_error
from typing import Callable, Any, List


class BaseRepository(object):
    __metaclass__ = ABCMeta

    def __init__(self, mbcontext):
        # type: (MBEasyContext) -> None
        self.mbcontext = mbcontext

    def execute_with_connection(self, method, db_name=None, service_name=None):
        # type: (Callable[[Connection], Any], int, str) -> Any
        conn = None
        try:
            conn = Driver().open(self.mbcontext,
                                 dbname=str(db_name) if db_name is not None else None,
                                 service_name=service_name)
            return method(conn)
        finally:
            if conn is not None:
                conn.close()

    def execute_with_transaction(self, method, db_name=None, service_name=None):
        # type: (Callable[[Connection], Any], int, str) -> Any
        def transaction_method(conn):
            # type: (Connection) -> Any

            conn.transaction_start()
            try:
                conn.query("BEGIN TRANSACTION")
                ret = method(conn)
                conn.query("COMMIT")
                return ret
            except:
                conn.query("ROLLBACK")
                raise
            finally:
                conn.transaction_end()

        return self.execute_with_connection(transaction_method, db_name=db_name, service_name=service_name)

    def execute_in_all_databases(self, method, pos_list):
        # type: (Callable[[Connection], Any], List[int]) -> Any
        ret = []
        for pos in pos_list:
            conn = None
            try:
                conn = Driver().open(self.mbcontext, dbname=str(pos))
                ret.append(method(conn))
            finally:
                if conn is not None:
                    conn.close()

        return ret

    def execute_in_all_databases_returning_flat_list(self, method, pos_list):
        # type: (Callable[[Connection], Any], List[int]) -> Any
        inner_ret = self.execute_in_all_databases(method, pos_list)

        ret = []
        for inner_list in inner_ret:
            ret.extend(inner_list)

        return ret


def show_message_dialog(pos_id, title="$INFORMATION", message="", mask="", assync=False):
    return show_any_dialog(pos_id, "messageDialog", message, title, "", "", "", 600000, mask, "", "", "", None, assync)


def show_message_options_dialog(pos_id, buttons, title="$INFORMATION", message=""):
    resp = show_any_dialog(
        pos_id, "messageOptionsDialog", message, title, "", buttons, "", 600000, "", "", "", "", None, False)

    if resp in (None, ''):
        return resp

    return int(resp)


def show_numpad_dialog(pos_id, title="", message="", mask="INTEGER", default_value=""):
    resp = show_keyboard(pos_id, message, title=title, mask=mask, numpad=True, defvalue=default_value)

    if resp in (None, '', 'NaN'):
        return resp

    mask = mask.lower()
    if mask == 'password':
        return str(resp)

    if mask != "currency" and mask != "number":
        return int(resp)

    return float(resp)


def show_filtered_list_box_dialog(pos_id, options, title="", message="", mask="", buttons="$OK|$CANCEL"):
    if type(options) == list:
        options = "|".join(options)
    resp = show_any_dialog(
        pos_id, "filteredListBox", message, title, options, buttons, "",  600000, mask, "", "", "", None, False)

    if resp in (None, ''):
        return resp

    return int(resp)


def show_check_list_box_dialog(pos_id, options, title="", message="", mask="", buttons="$OK|$CANCEL"):
    if type(options) == list:
        options = "|".join(options)
    resp = show_any_dialog(
        pos_id, "checkListBox", message, title, options, buttons, "",  600000, mask, "", "", "", None, False)

    if resp in (None, ''):
        return []

    resp = resp.split(';')
    return resp


def show_editable_list_box_dialog(pos_id, options="", title="", buttons="$OK|$CANCEL"):
    resp = show_any_dialog(
        pos_id, "multipleColumnListBox", "", title, options, buttons, "",  600000, "", "", "", "", None, False)

    return resp

def show_bordereau_dialog(pos_id):
    return show_any_dialog(
        pos_id, "bordereauDialog", '', "$BORDEREAU", "", "", "", 600000, "mask", "", "", "", None, False)


def show_confirm_dialog(pos_id, title=''):
    if not title:
        title = "$BORDEREAU"
    return show_any_dialog(
        pos_id, "bordereauDialog", 'confirmDialog', title, "", "", "", 600000, "mask", "", "", "", None, False)


def components_is_running(mb_context, *params):
    msg_resp = mb_context.MB_EasySendMessage("HyperVisor", TK_HV_COMPLIST, FM_PARAM)
    if msg_resp.token != TK_SYS_ACK:
        sys_log_error("Unable to obtain the TK_HV_COMPLIST")
        return False

    component_list = [_.get('name').lower() for _ in eTree.XML(msg_resp.data).findall(".//component[@status='1']")]
    for component in params:
        if component.lower() not in component_list:
            return False

    return True


def ensure_iterable(value):
    if isinstance(value, basestring):
        return [value]

    try:
        _ = (e for e in value)
        return value
    except TypeError:
        return [value]


def ensure_list(value):
    if isinstance(value, basestring):
        return [value]

    if isinstance(value, list):
        return value

    try:
        _ = (e for e in value)
        return list(value)
    except TypeError:
        return [value]


def convert_to_dict(iterable):
    ret = {}
    for item in iterable:
        ret[item] = item
    return ret


def convert_from_localtime_to_utc(localtime):
    # type: (datetime) -> datetime
    from_zone = tz.tzlocal()
    to_zone = tz.tzutc()
    localtime = localtime.replace(tzinfo=from_zone)
    utc = localtime.astimezone(to_zone)

    return utc



# -*- coding: utf-8 -*-
import os

from helper import config_logger, import_pydevd
from old_helper import get_class_by_name
from messagehandler import MessageHandler
from msgbus import MBEasyContext
from pos_model import OrderParser

from eventhandler import NTaxCalcEventHandler
from taxcalculator import TaxCalculatorService, GeneralTaxCalculator, OrderFormatter, TaxItemFormatter
from taxcalculator.repository import NTaxCalcRepository

REQUIRED_SERVICES = ""


def main():
    import_pydevd(os.environ["LOADERCFG"], 9175)

    config_logger(os.environ["LOADERCFG"], 'NTaxCalc')
    mbcontext = MBEasyContext("NTaxCalc")

    message_handler = MessageHandler(mbcontext, "NTaxCalc", "NTaxCalc", REQUIRED_SERVICES)

    order_parser = OrderParser()

    tax_calculators = []
    ntaxcalc_repository = NTaxCalcRepository(mbcontext)

    taxes = ntaxcalc_repository.load_all_taxes()
    for tax in taxes:
        tax_products = ntaxcalc_repository.load_all_products_from_tax(tax.code)

        klz = get_class_by_name(tax.tax_processor)
        tax_calculator = klz(mbcontext, tax.code, tax.name, tax.fiscal_index, tax.rate, tax_products, tax.params)

        tax_calculators.append(tax_calculator)

    general_tax_calculator = GeneralTaxCalculator(tax_calculators)

    tax_item_formatter = TaxItemFormatter()
    order_formatter = OrderFormatter(tax_item_formatter)
    tax_calculator_service = TaxCalculatorService(order_parser, general_tax_calculator, order_formatter)

    event_handler = NTaxCalcEventHandler(mbcontext, tax_calculator_service)

    message_handler.set_event_handler(event_handler)

    message_handler.subscribe_sync_events(["TAXCALC", "TAX_CALC"])

    message_handler.handle_events()



from typing import Dict

from ._Route import Route


class FoundRoute(object):
    def __init__(self, route, path_parameters):
        # type: (Route, Dict[str, str]) -> None
        self.route = route
        self.path_parameters = path_parameters



import os

import cfgtools
from externals.events import ListenedReentrantEvents, ListenedNonReentrantEvents
from externals.message_handler_builder import HvMessageHandlerBuilder
from externals.tokens import ListenedQueueTokens
from helper import import_pydevd
from mbcontextmessagehandler import MbContextMessageBus, MbContextMessageHandler
from msgbus import MBEasyContext
from startup.common_main import configure_logger, COMPONENT_NAME
from typing import Dict, List


def main():
    # type: () -> None

    loader_path = os.environ["LOADERCFG"]
    import_pydevd(loader_path, 9160)

    reader = cfgtools.read(loader_path)
    component_name = str(reader.find_value("Identification.ComponentName", default=COMPONENT_NAME))

    device_configurations = {key.name: key.value() if key.type != 1 else key.values
                             for key in reader.find_group("NPeripherals").keys}

    service_names = _get_service_names(device_configurations)

    logger = configure_logger(component_name, service_names[0])

    mb_context = MBEasyContext(component_name)

    message_bus = MbContextMessageBus(mb_context)

    hv_message_handler_builder = HvMessageHandlerBuilder(message_bus, device_configurations)
    message_handler = MbContextMessageHandler(message_bus, "", "", "", hv_message_handler_builder)

    mb_context.MB_EasyWaitStart()

    _register_services(component_name, mb_context, service_names)

    _register_non_reentrant_events(message_handler)

    _register_reentrant_events(message_handler)

    _register_queued_tokens(message_handler)

    logger.debug("Starting hv [{}][{}] ...".format(component_name, service_names))
    message_handler.handle_events()

    mb_context.MB_EasyFinalize()


def _register_non_reentrant_events(message_handler):
    listened_non_reentrant_events = ListenedNonReentrantEvents.list_names()
    if listened_non_reentrant_events:
        message_handler.subscribe_non_reentrant_events(listened_non_reentrant_events)


def _register_reentrant_events(message_handler):
    listened_reentrant_events = ListenedReentrantEvents.list_names()
    if listened_reentrant_events:
        message_handler.subscribe_reentrant_events(listened_reentrant_events)


def _register_queued_tokens(message_handler):
    listened_queue_tokens = ListenedQueueTokens.list_values()
    if listened_queue_tokens:
        message_handler.subscribe_queue_messages(listened_queue_tokens)


def _register_services(component_name, mb_context, service_names):
    for service in service_names:
        mb_context.MB_EasyRegisterService(component_name, service)


def _get_service_names(device_configurations):
    # type: (Dict[str, str]) -> List[str]
    service_names = device_configurations.get("Name", None)
    if not isinstance(service_names, list):
        return [service_names]
    return service_names


if __name__ == "__main__":
    main()



from helper import ExtendedEnum


class MwSysErrors(ExtendedEnum):

    SE_UNKOWNUSR = "5000"         # unknown user id
    SE_INVALIDPASSWD = "5001"     # invalid password
    SE_DAYNOTCLOSED = "5002"      # business day is not closed
    SE_DAYNOTOPENED = "5003"      # business day is not opened
    SE_OPERLOGGEDIN = "5004"      # operator is still logged in
    SE_OPERNOTLOGGEDIN = "5005"   # operator is not logged in
    SE_ORDEROPENED = "5006"       # there are a sale order opened
    SE_OPEROPENED = "5007"        # there are operators opened in the account database



from six import PY3

from functools import wraps

from datetime import datetime, timedelta, tzinfo


ZERO = timedelta(0)

__all__ = ['tzname_in_python2', 'enfold']


def tzname_in_python2(namefunc):
    """Change unicode output into bytestrings in Python 2

    tzname() API changed in Python 3. It used to return bytes, but was changed
    to unicode strings
    """
    def adjust_encoding(*args, **kwargs):
        name = namefunc(*args, **kwargs)
        if name is not None and not PY3:
            name = name.encode()

        return name

    return adjust_encoding


# The following is adapted from Alexander Belopolsky's tz library
# https://github.com/abalkin/tz
if hasattr(datetime, 'fold'):
    # This is the pre-python 3.6 fold situation
    def enfold(dt, fold=1):
        """
        Provides a unified interface for assigning the ``fold`` attribute to
        datetimes both before and after the implementation of PEP-495.

        :param fold:
            The value for the ``fold`` attribute in the returned datetime. This
            should be either 0 or 1.

        :return:
            Returns an object for which ``getattr(dt, 'fold', 0)`` returns
            ``fold`` for all versions of Python. In versions prior to
            Python 3.6, this is a ``_DatetimeWithFold`` object, which is a
            subclass of :py:class:`datetime.datetime` with the ``fold``
            attribute added, if ``fold`` is 1.

        .. versionadded:: 2.6.0
        """
        return dt.replace(fold=fold)

else:
    class _DatetimeWithFold(datetime):
        """
        This is a class designed to provide a PEP 495-compliant interface for
        Python versions before 3.6. It is used only for dates in a fold, so
        the ``fold`` attribute is fixed at ``1``.

        .. versionadded:: 2.6.0
        """
        __slots__ = ()

        @property
        def fold(self):
            return 1

    def enfold(dt, fold=1):
        """
        Provides a unified interface for assigning the ``fold`` attribute to
        datetimes both before and after the implementation of PEP-495.

        :param fold:
            The value for the ``fold`` attribute in the returned datetime. This
            should be either 0 or 1.

        :return:
            Returns an object for which ``getattr(dt, 'fold', 0)`` returns
            ``fold`` for all versions of Python. In versions prior to
            Python 3.6, this is a ``_DatetimeWithFold`` object, which is a
            subclass of :py:class:`datetime.datetime` with the ``fold``
            attribute added, if ``fold`` is 1.

        .. versionadded:: 2.6.0
        """
        if getattr(dt, 'fold', 0) == fold:
            return dt

        args = dt.timetuple()[:6]
        args += (dt.microsecond, dt.tzinfo)

        if fold:
            return _DatetimeWithFold(*args)
        else:
            return datetime(*args)


def _validate_fromutc_inputs(f):
    """
    The CPython version of ``fromutc`` checks that the input is a ``datetime``
    object and that ``self`` is attached as its ``tzinfo``.
    """
    @wraps(f)
    def fromutc(self, dt):
        if not isinstance(dt, datetime):
            raise TypeError("fromutc() requires a datetime argument")
        if dt.tzinfo is not self:
            raise ValueError("dt.tzinfo is not self")

        return f(self, dt)

    return fromutc


class _tzinfo(tzinfo):
    """
    Base class for all ``dateutil`` ``tzinfo`` objects.
    """

    def is_ambiguous(self, dt):
        """
        Whether or not the "wall time" of a given datetime is ambiguous in this
        zone.

        :param dt:
            A :py:class:`datetime.datetime`, naive or time zone aware.


        :return:
            Returns ``True`` if ambiguous, ``False`` otherwise.

        .. versionadded:: 2.6.0
        """

        dt = dt.replace(tzinfo=self)

        wall_0 = enfold(dt, fold=0)
        wall_1 = enfold(dt, fold=1)

        same_offset = wall_0.utcoffset() == wall_1.utcoffset()
        same_dt = wall_0.replace(tzinfo=None) == wall_1.replace(tzinfo=None)

        return same_dt and not same_offset

    def _fold_status(self, dt_utc, dt_wall):
        """
        Determine the fold status of a "wall" datetime, given a representation
        of the same datetime as a (naive) UTC datetime. This is calculated based
        on the assumption that ``dt.utcoffset() - dt.dst()`` is constant for all
        datetimes, and that this offset is the actual number of hours separating
        ``dt_utc`` and ``dt_wall``.

        :param dt_utc:
            Representation of the datetime as UTC

        :param dt_wall:
            Representation of the datetime as "wall time". This parameter must
            either have a `fold` attribute or have a fold-naive
            :class:`datetime.tzinfo` attached, otherwise the calculation may
            fail.
        """
        if self.is_ambiguous(dt_wall):
            delta_wall = dt_wall - dt_utc
            _fold = int(delta_wall == (dt_utc.utcoffset() - dt_utc.dst()))
        else:
            _fold = 0

        return _fold

    def _fold(self, dt):
        return getattr(dt, 'fold', 0)

    def _fromutc(self, dt):
        """
        Given a timezone-aware datetime in a given timezone, calculates a
        timezone-aware datetime in a new timezone.

        Since this is the one time that we *know* we have an unambiguous
        datetime object, we take this opportunity to determine whether the
        datetime is ambiguous and in a "fold" state (e.g. if it's the first
        occurence, chronologically, of the ambiguous datetime).

        :param dt:
            A timezone-aware :class:`datetime.datetime` object.
        """

        # Re-implement the algorithm from Python's datetime.py
        dtoff = dt.utcoffset()
        if dtoff is None:
            raise ValueError("fromutc() requires a non-None utcoffset() "
                             "result")

        # The original datetime.py code assumes that `dst()` defaults to
        # zero during ambiguous times. PEP 495 inverts this presumption, so
        # for pre-PEP 495 versions of python, we need to tweak the algorithm.
        dtdst = dt.dst()
        if dtdst is None:
            raise ValueError("fromutc() requires a non-None dst() result")
        delta = dtoff - dtdst

        dt += delta
        # Set fold=1 so we can default to being in the fold for
        # ambiguous dates.
        dtdst = enfold(dt, fold=1).dst()
        if dtdst is None:
            raise ValueError("fromutc(): dt.dst gave inconsistent "
                             "results; cannot convert")
        return dt + dtdst

    @_validate_fromutc_inputs
    def fromutc(self, dt):
        """
        Given a timezone-aware datetime in a given timezone, calculates a
        timezone-aware datetime in a new timezone.

        Since this is the one time that we *know* we have an unambiguous
        datetime object, we take this opportunity to determine whether the
        datetime is ambiguous and in a "fold" state (e.g. if it's the first
        occurance, chronologically, of the ambiguous datetime).

        :param dt:
            A timezone-aware :class:`datetime.datetime` object.
        """
        dt_wall = self._fromutc(dt)

        # Calculate the fold status given the two datetimes.
        _fold = self._fold_status(dt, dt_wall)

        # Set the default fold value for ambiguous dates
        return enfold(dt_wall, fold=_fold)


class tzrangebase(_tzinfo):
    """
    This is an abstract base class for time zones represented by an annual
    transition into and out of DST. Child classes should implement the following
    methods:

        * ``__init__(self, *args, **kwargs)``
        * ``transitions(self, year)`` - this is expected to return a tuple of
          datetimes representing the DST on and off transitions in standard
          time.

    A fully initialized ``tzrangebase`` subclass should also provide the
    following attributes:
        * ``hasdst``: Boolean whether or not the zone uses DST.
        * ``_dst_offset`` / ``_std_offset``: :class:`datetime.timedelta` objects
          representing the respective UTC offsets.
        * ``_dst_abbr`` / ``_std_abbr``: Strings representing the timezone short
          abbreviations in DST and STD, respectively.
        * ``_hasdst``: Whether or not the zone has DST.

    .. versionadded:: 2.6.0
    """
    def __init__(self):
        raise NotImplementedError('tzrangebase is an abstract base class')

    def utcoffset(self, dt):
        isdst = self._isdst(dt)

        if isdst is None:
            return None
        elif isdst:
            return self._dst_offset
        else:
            return self._std_offset

    def dst(self, dt):
        isdst = self._isdst(dt)

        if isdst is None:
            return None
        elif isdst:
            return self._dst_base_offset
        else:
            return ZERO

    @tzname_in_python2
    def tzname(self, dt):
        if self._isdst(dt):
            return self._dst_abbr
        else:
            return self._std_abbr

    def fromutc(self, dt):
        """ Given a datetime in UTC, return local time """
        if not isinstance(dt, datetime):
            raise TypeError("fromutc() requires a datetime argument")

        if dt.tzinfo is not self:
            raise ValueError("dt.tzinfo is not self")

        # Get transitions - if there are none, fixed offset
        transitions = self.transitions(dt.year)
        if transitions is None:
            return dt + self.utcoffset(dt)

        # Get the transition times in UTC
        dston, dstoff = transitions

        dston -= self._std_offset
        dstoff -= self._std_offset

        utc_transitions = (dston, dstoff)
        dt_utc = dt.replace(tzinfo=None)

        isdst = self._naive_isdst(dt_utc, utc_transitions)

        if isdst:
            dt_wall = dt + self._dst_offset
        else:
            dt_wall = dt + self._std_offset

        _fold = int(not isdst and self.is_ambiguous(dt_wall))

        return enfold(dt_wall, fold=_fold)

    def is_ambiguous(self, dt):
        """
        Whether or not the "wall time" of a given datetime is ambiguous in this
        zone.

        :param dt:
            A :py:class:`datetime.datetime`, naive or time zone aware.


        :return:
            Returns ``True`` if ambiguous, ``False`` otherwise.

        .. versionadded:: 2.6.0
        """
        if not self.hasdst:
            return False

        start, end = self.transitions(dt.year)

        dt = dt.replace(tzinfo=None)
        return (end <= dt < end + self._dst_base_offset)

    def _isdst(self, dt):
        if not self.hasdst:
            return False
        elif dt is None:
            return None

        transitions = self.transitions(dt.year)

        if transitions is None:
            return False

        dt = dt.replace(tzinfo=None)

        isdst = self._naive_isdst(dt, transitions)

        # Handle ambiguous dates
        if not isdst and self.is_ambiguous(dt):
            return not self._fold(dt)
        else:
            return isdst

    def _naive_isdst(self, dt, transitions):
        dston, dstoff = transitions

        dt = dt.replace(tzinfo=None)

        if dston < dstoff:
            isdst = dston <= dt < dstoff
        else:
            isdst = not dstoff <= dt < dston

        return isdst

    @property
    def _dst_base_offset(self):
        return self._dst_offset - self._std_offset

    __hash__ = None

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return "%s(...)" % self.__class__.__name__

    __reduce__ = object.__reduce__


def _total_seconds(td):
    # Python 2.6 doesn't have a total_seconds() method on timedelta objects
    return ((td.seconds + td.days * 86400) * 1000000 +
            td.microseconds) // 1000000


_total_seconds = getattr(timedelta, 'total_seconds', _total_seconds)
